#include <avr/io.h>
#include <avr/pgmspace.h>
#ifndef F_CPU
#error "F_CPU nicht definiert"
#endif
#include <util/delay.h>
#include "lcd_drv.h"

//-----------------------------------------------------------------------------
void lcd_nibble( uint8_t d )
{
  LCD_D4_PORT &= ~(1<<LCD_D4);
  if( d & 1<<4 ) LCD_D4_PORT |= (1<<LCD_D4);

  LCD_D5_PORT &= ~(1<<LCD_D5);
  if( d & 1<<5 ) LCD_D5_PORT |= (1<<LCD_D5);

  LCD_D6_PORT &= ~(1<<LCD_D6);
  if( d & 1<<6 ) LCD_D6_PORT |= (1<<LCD_D6);

  LCD_D7_PORT &= ~(1<<LCD_D7);
  if( d & 1<<7 ) LCD_D7_PORT |= (1<<LCD_D7);

  LCD_E_PORT  |=  (1<<LCD_E );
  _delay_us( 1 );              // wait 1us
  LCD_E_PORT  &= ~(1<<LCD_E );

}
//-----------------------------------------------------------------------------
void lcd_byte( uint8_t d )
{
  lcd_nibble( d    );
  lcd_nibble( d<<4 );
  _delay_us( 45   );           // wait 45us
}
//-----------------------------------------------------------------------------
void lcd_command( uint8_t d )
{
  LCD_RS_PORT &= ~(1<<LCD_RS);
  lcd_byte( d );
  switch( d ){
    case 1:
    case 2:
    case 3: _delay_ms( 2 );    // wait 2ms
  }
}
//-----------------------------------------------------------------------------
void lcd_data( uint8_t d )
{
  LCD_RS_PORT |= (1<<LCD_RS);
  lcd_byte( d );
}
//-----------------------------------------------------------------------------
void lcd_init( void ) {
/* Bereits in main() erledigt!
  LCD_D4_DDR |=  (1<<LCD_D4);
  LCD_D5_DDR |=  (1<<LCD_D5);
  LCD_D6_DDR |=  (1<<LCD_D6);
  LCD_D7_DDR |=  (1<<LCD_D7);
  LCD_RS_DDR |=  (1<<LCD_RS);
  LCD_E_DDR  |=  (1<<LCD_E );

  LCD_E_PORT  &= ~(1<<LCD_E );
  LCD_RS_PORT &= ~(1<<LCD_RS); // send commands
*/
  _delay_ms( 15 );             // wait 15ms

  lcd_nibble( 0x30 );
  _delay_ms( 5);               // wait >4.1ms

  lcd_nibble( 0x30 );
  _delay_us( 100 );            // wait >100us

  lcd_nibble( 0x30 );          // 8 bit mode
  _delay_us( 100 );            // wait >100us

  lcd_nibble( 0x20 );          // 4 bit mode
  _delay_us( 100 );            // wait >100us

  lcd_command( 0x28 );         // 2 lines 5*7
  lcd_command( 0x08 );         // display off
  lcd_command( 0x01 );         // display clear
  lcd_command( 0x06 );         // cursor increment

  lcd_command (0x40); // Zeichen #0 wird Reglersymbol
  lcd_data (0x00);     // Zeile 1: ........
  lcd_data (0x18);     // Zeile 2: ...**...
  lcd_data (0x10);     // Zeile 3:  ...*...
  lcd_data (0x04);     // Zeile 4: ..*.....
  lcd_data (0x01);     // Zeile 5: :*......
  lcd_data (0x03);     // Zeile 6: :**.....
  lcd_data (0x00);     // Zeile 7: :.......
  lcd_data (0x00);     // Zeile 8: :.......

  lcd_command( 0x0C );         // on, no cursor, no blink
}
//-----------------------------------------------------------------------------
void lcd_pos( uint8_t line, uint8_t column )
{
  switch( line )		        // Zeile 1: 1, Spalte 1: 1
  {
    case 2: column += 0x40; break;
    case 3: column += 0x14; break;
    case 4: column += 0x54; break;
  }

  lcd_command( 0x7F + column );
}
//-----------------------------------------------------------------------------
void lcd_text( char *t ) {
  while( *t )  {
    lcd_data( *t );
    t++;
  }
}
/*-----------------------------------------------------------------------------
 schreibt die aus dem Programmspeicher bergebene Zeichenkette
 an die aktuelle Cursor-Position.
 Def. der Zeichenkette pgm_s vor main(): const char pgm_s[] PROGMEM = "ABCD";
 Def. der Zeichenkette pgm_s in main():  static const char pgm_s[] PROGMEM =
 oder : lcd_text_pgm(PSTR("XYZ"));
------------------------------------------------------------------------------*/
void lcd_text_pgm(const char *progmem_string) {
  register char c;
  while ((c = pgm_read_byte(progmem_string++)))		lcd_data(c);
}

/*-----------------------------------------------------------------------------
 Zahl auf LCD ausgeben: bin: Zahl<65 536,      digit: Stellenzahl ohne Komma
                        dez: Kommastellen,     space: Ohne fuehrende NULL
-----------------------------------------------------------------------------*/
void lcd_uint(uint16_t bin , uint8_t digit , uint8_t dez , uint8_t space) {
  uint8_t  i;
  uint16_t sub;
// Tabelle der Zehnerpotenzwerte:
  static const uint16_t TEST[4] PROGMEM = {    10,
                                              100,
                                             1000,
                                            10000,};

  if(digit>5) digit=5;               // Zehnerpotenzueberlauf vermeiden
  char lcd[digit];
  if(digit<5) {                      // Ueberlauf der Anzeigestellen -> 999...
     sub=pgm_read_word(&TEST[digit-1])-1;
      if(bin > sub ) bin = sub;
  }
  for(i=digit-1;i>0;i--) {
      sub=pgm_read_word(&TEST[i-1]);
      lcd[i]=0;
      while(bin>=sub) {
        bin -=sub;                    // Zahl - Zehnerpotenz
        lcd[i]++;                     // Anzahl der Subtraktionen
      }
  }
  lcd[0] = (char)bin;                 // Einerstelle
  if(digit == dez) {                  // Ziffern ausgeben
      lcd_data('0');
      lcd_data('.');
  }
  while(digit) {
    digit--;
    if(space&&(digit>dez)&&(lcd[digit]==0)) lcd_data(' ');
    else  {
      lcd_data(lcd[digit]+'0');       // Ziffer in ASCI umwandeln und an LCD
      space=0;                        // ausgeben. Ab jetzt auch '0' ausgeben.
    }
    if((digit == dez)&&dez) lcd_data('.');
  }
}
