/*************** Frequenznormal, Beschaltung ATMEGA328P *********************
PB0            ICP1/GPS_Takt | PC0      LCD_RS    | PD0/RXD    RX
PB1/OC1A       SW            | PC1      LCD_E     | PD1/TXD    TX
PB2/OC1B       AD5683_SYNC   | PC2      LCD_D4    | PD2/INT0   STROMAUSFALL
PB3/MOSI       AD5683_MOSI   | PC3      LCD_D5    | PD3/OC2B   TAKTAUSGANG_2
PB4/MISO       FREI          | PC4/SDA  LCD_D6    | PD4        FREI
PB5/ SCK       AD5683_SCK    | PC5/SCL  LCD_D7    | PD5/OC0B   TAKTAUSGANG_1
PB6/XTAL1      TCXO_20MHz    | ADC6     FREI      | PD6/OC0A   ENCODER
PB7/XTAL2      FREI          | ADC7     FREI      | PD7        ENCODER
                             | PC6      RESET     |
****************************** Takt: 20MHz  *******************************/

#include <avr/io.h>
#ifndef F_CPU
#error "F_CPU nicht definiert"
#endif
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>
#include "lcd_drv.h"
#include "AD5683.h"

#define BAUD 38400UL                          // Baudrate
// Berechnungen
#define UBRR_VAL ((F_CPU+BAUD*4)/(BAUD*8)-1)  // U2X0=1, clever runden
#define BAUD_REAL (F_CPU/(8*(UBRR_VAL+1)))    // U2X0=1,Reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD)    // Fehler in 0,1%, 1000 = 0,00%
#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
#error Baudrateabweichung grsser 1%!
#endif

#define DAC 43400U      // DAC Nominalwert
#define DAC_DELTA 55    // DAC-Abweichung bei einem Fehler von 0,5 Takte/s
// DAC AD5683: 110 digit pro Hz (gemessen)
// DAC-Wert 17000 (0,65V): Rest 47259 Takte (Minimum -12ppm) bei Messdauer 10s
// DAC-Wert 43332 (1,65V): Rest 49664 Takte (Sollwert  0ppm) bei Messdauer 10s
// DAC-Wert 65535 (2,50V): Rest 51689 Takte (Maximum +10ppm) bei Messdauer 10s
#define REST1  11520U    // REST1  =  20000000Takte - ( 305 * 2e16Takte)
#define REST10 49664U    // REST10 = 200000000Takte - (3051 * 2e16Takte)

uint16_t dac_wert;
uint16_t ee_dac_wert EEMEM;
volatile uint16_t icr, t1_ovf;

uint8_t gga_time[6], gga_latitude[10], gga_n, gga_longitude[11], gga_e,
        gga_fs, gga_nosv[2], gga_msl[5];

volatile uint8_t ocr0a, ocr0b, cnt_max0, ocr2a, ocr2b, cnt_max2;
int8_t takt0, takt2, time_zone, taktq ;
uint8_t ee_takt0 EEMEM, ee_takt2 EEMEM, ee_time_zone EEMEM;

static const uint16_t PROGMEM freq_a[29]={5,10,20,40,50,80,100,200,400,500,
                               1,2,4,5,8,10,16,20,40,50,80,100,200,400,500,
                               1,2,5,10};

// f = 20MHz/(2*clk_Div*(((n-1)*(OCRnA+1))+(OCRnB+1))
// oder bei n=0: f = 20MHz/(2*clk_Div*(OCRnA+1))
// mit clk_Div=1 bei CSnx=1, =8 bei CSnx=2, =64 bei CSnx=3
static const uint8_t PROGMEM freq_p[29][5]={ // pgm_read_byte(&freq_p[n][m]));
  {247,249,3,126,' '}, // OCRnA, OCRnB, CSnx, n-Mal      5Hz   0
  {251,252,3, 62,' '}, // OCRnA, OCRnB, CSnx, n-Mal     10Hz   1
  {248,249,2,251,' '}, // OCRnA, OCRnB, CSnx, n-Mal     20Hz   2
  {251,253,2,124,' '}, // OCRnA, OCRnB, CSnx, n-Mal     40Hz   3
  {237,247,2,105,' '}, // OCRnA, OCRnB, CSnx, n-Mal     50Hz   4
  {251,252,2, 62,' '}, // OCRnA, OCRnB, CSnx, n-Mal     80Hz   5
  {244,249,2, 51,' '}, // OCRnA, OCRnB, CSnx, n-Mal    100Hz   6
  {239,249,2, 26,' '}, // OCRnA, OCRnB, CSnx, n-Mal    200Hz   7
  {239,244,2, 13,' '}, // OCRnA, OCRnB, CSnx, n-Mal    400Hz   8
  {226,229,2, 11,' '}, // OCRnA, OCRnB, CSnx, n-Mal    500Hz   9
  {248,253,2,  5,'k'}, // OCRnA, OCRnB, CSnx, n-Mal     1kHz  10
  {184,254,2,  3,'k'}, // OCRnA, OCRnB, CSnx, n-Mal     2kHz  11
  {226,229,1, 11,'k'}, // OCRnA, OCRnB, CSnx, n-Mal     4kHz  12
  {249,  0,2,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal     5kHz  13
  {248,253,1,  5,'k'}, // OCRnA, OCRnB, CSnx, n-Mal     8kHz  14
  {124,  0,2,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal    10kHz  15
  {184,254,1,  3,'k'}, // OCRnA, OCRnB, CSnx, n-Mal    16kHz  16
  {243,255,1,  2,'k'}, // OCRnA, OCRnB, CSnx, n-Mal    20kHz  17
  {249,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal    40kHz  18
  {199,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal    50kHz  19
  {124,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal    80kHz  20
  { 99,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal   100kHz  21
  { 49,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal   200kHz  22
  { 24,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal   400kHz  23
  { 19,  0,1,  0,'k'}, // OCRnA, OCRnB, CSnx, n-Mal   500kHz  24
  {  9,  0,1,  0,'M'}, // OCRnA, OCRnB, CSnx, n-Mal     1MHz  25
  {  4,  0,1,  0,'M'}, // OCRnA, OCRnB, CSnx, n-Mal     2MHz  26
  {  1,  0,1,  0,'M'}, // OCRnA, OCRnB, CSnx, n-Mal     5MHz  27
  {  0,  0,1,  0,'M'}, // OCRnA, OCRnB, CSnx, n-Mal    10MHz  28

};
///////////////////////// Flags zur Ablaufsteuerung ///////////////////////////
#define FLAG           GPIOR0

#define TIMEPULSE      (1<<0)   // 1x pro Sekunde bei Empfang des GPS-Signals
#define NO_GPS         (1<<1)   // H, wenn Sekundenimpuls nicht vorhanden
#define IN_SYNC        (1<<2)   // Max. +-1Takt/10s Abweichung zu Sekundenimpuls
#define OLD_SYNC_VAL   (1<<3)   // Sync-Zustand der vergangenen Messperiode
#define T_ZONE         (1<<4)   // Einstellmodus fuer Zeitzone
#define MENU           (1<<5)   // Einstellmodus fuer Zeitzone oder Ausgaenge
#define TAKTQ          (1<<6)   // Einstellmodus fuer gewaehlten Taktausgang
#define WAHL           (1<<7)   // Wahl des Einstellmodus fuer Taktausgang 1/2
uint8_t reset;                  // Reset der Statistik
////////////////////////////////////// Encoder ////////////////////////////////
#define PHASE_A   (PIND & 1<<PD7)
#define PHASE_B   (PIND & 1<<PD6)
volatile int8_t enc_delta;       // -128 ... 127
int8_t last;
/////////////////////////////////// Tastenentprellung /////////////////////////
#define KEY_DDR       DDRB
#define KEY_PORT      PORTB
#define KEY_PIN       PINB
#define KEY0          1
#define ALL_KEYS      (1<<KEY0)

#define REPEAT_MASK   (1<<KEY0)  // repeat: key0
#define REPEAT_START  30         // after 600ms
// volatile uint8_t key_state;   // debounced and inverted key state:
                                 // bit = 1: key pressed
// volatile uint8_t key_press;   // key press detect
// volatile uint8_t key_rpt;     // key long press and repeat

#define key_state     GPIOR1      // debounced and inverted key state:
                                 // bit = 1: key pressed
#define key_press     GPIOR2     // key press detect
#define key_rpt       PCMSK0     // key long press and repeat

//////////////////////////////////// Encoder //////////////////////////////////
static void encode_init( void )  {
//  Pollin: PHASE_A > Anschluss B und PHASE_B > Anschluss A
//  Falls die internen Pullups verwendet werden, ist nach deren Initialisierung
//  eine kurze Pause erforderlich
//  _delay_ms(1);
  int8_t neu;
  if( PHASE_A )
    neu = 3;
  if( PHASE_B )
    neu ^= 1;                  // convert gray to binary
  last = neu;                  // power on state
//Pollin Encoder: Zhlwerk auf stabile Startposition 0/1 setzen:
  enc_delta = ( (neu&1) ? 0 : 1 );
//  enc_delta = 0;             // Herkoemliche Version
}

int8_t encode_read( void ) {
  int8_t val;
  cli();
  val = enc_delta;
   enc_delta = val & 1;    // read two step encoders
//  enc_delta = val & 3;       // read four step encoders
  sei();
  return val >> 1;        // read two step encoders
  // return val >> 2;           // read four step encoders
}
//////////////////////////// Tastenentprellung ////////////////////////////////
//
// check if a key has been pressed. Each pressed key is reported
// only once
//
uint8_t get_key_press( uint8_t key_mask ) {
  cli();                                          // read and clear atomic !
  key_mask &= key_press;                          // read key(s)
  key_press ^= key_mask;                          // clear key(s)
  sei();
  return key_mask;
}
// ---------------------------------------------------------------------------
// check if a key has been pressed long enough such that the
// key repeat functionality kicks in. After a small setup delay
// the key is reported beeing pressed in subsequent calls
// to this function. This simulates the user repeatedly
// pressing and releasing the key.

uint8_t get_key_rpt( uint8_t key_mask ) {
  cli();                                          // read and clear atomic !
  key_mask &= key_rpt;                            // read key(s)
  key_rpt ^= key_mask;                            // clear key(s)
  sei();
  return key_mask;
}
// ---------------------------------------------------------------------------
uint8_t get_key_short( uint8_t key_mask ) {
  cli();                               // read key state and key press atomic!
  return get_key_press( ~key_state & key_mask );
}
// ---------------------------------------------------------------------------
uint8_t get_key_long( uint8_t key_mask ) {
  return get_key_press( get_key_rpt( key_mask ));
}
///////////////////////////// LCD loeschen ////////////////////////////////////
void lcd_clz(void) {for(uint8_t i = 0 ; i < 20; i++ ) lcd_data(' ');}
// ----------------------------------------------------------------------------
void lcd_clear(void) {
  lcd_pos(1,1);
  lcd_text_pgm(PSTR("* Kein GPS-Empfang *"));
  lcd_pos(2,1); lcd_clz();
  lcd_pos(3,1); lcd_clz();
}
//////////////////////////////// read_EEPROM() ////////////////////////////////
// ee_takt0 EEMEM, ee_takt2 EEMEM, ee_time_zone
static void read_EEPROM(void) {  // Uebernahme des DAC-Wertes aus dem EEPROM
  uint16_t tmp;
  uint8_t  tmp1;
  tmp = eeprom_read_word(&ee_dac_wert);
  if(tmp < 0xFFFF)  dac_wert = tmp; // Nur falls EEPROM nicht geloescht(0xFFFF)
  else dac_wert = DAC;              // 1,65V

  tmp1 = eeprom_read_byte(&ee_takt0);
  if(tmp1 < 0xFF)  takt0 = tmp1;    // Nur falls EEPROM nicht geloescht(0xFFFF)


  tmp1 = eeprom_read_byte(&ee_takt2);
  if(tmp1 < 0xFF) takt2 = tmp1;     // Nur falls EEPROM nicht geloescht(0xFFFF)

  tmp1 = eeprom_read_byte(&ee_time_zone);
  if(tmp1 < 0xFF) time_zone = tmp1; // Nur falls EEPROM nicht geloescht(0xFFFF)
}
////////////////// write_EEPROM() bei Stromausfall ////////////////////////////
// Speicherung des DAC-Wertes, der Zeitzone und der Takteinstellungen im EEPROM
void write_EEPROM(void) { 
  uint16_t tmp;
  uint8_t  tmp1;
  tmp = eeprom_read_word(&ee_dac_wert);
  if((tmp > (dac_wert + 20)) || (tmp < (dac_wert - 20))) {
    eeprom_write_word(&ee_dac_wert, dac_wert);
  }
  tmp1 = eeprom_read_byte(&ee_time_zone);
  if(ee_time_zone != time_zone) {
    eeprom_write_byte(&ee_time_zone, time_zone);
  }
  tmp1 = eeprom_read_byte(&ee_time_zone);
  if(tmp1 != time_zone) {
    eeprom_write_byte(&ee_time_zone, time_zone);
  }
  tmp1 = eeprom_read_byte(&ee_takt0);
  if(tmp1 != takt0) {
    eeprom_write_byte(&ee_takt0, takt0);    
  }
  tmp1 = eeprom_read_byte(&ee_takt2);
  if(tmp1 != takt2) {
    eeprom_write_byte(&ee_takt2, takt2);    
  }
}
//////////////////////////////// write_DAC() //////////////////////////////////
void write_DAC(void) {
  dac_write(DAC_WRITE, dac_wert);
  lcd_pos(3,9);
  lcd_data(0); // Reglersymbol
  if(dac_wert < DAC) { lcd_data('-'); lcd_uint((DAC - dac_wert) , 3 , 0, 0 ); }
  else               { lcd_data('+'); lcd_uint((dac_wert - DAC) , 3 , 0, 0 ); }

}
/////////////////////// Konfiguration des ATMEGA328P //////////////////////////
void cfg(void) {
  PORTB = 1<<PB0|1<<PB1|1<<PB2|0<<PB3|1<<PB4|0<<PB5|0<<PB6|1<<PB7;
  DDRB  = 0<<PB0|0<<PB1|1<<PB2|1<<PB3|0<<PB4|1<<PB5|0<<PB6|0<<PB7;

  PORTC = 0<<PC0|0<<PC1|0<<PC2|0<<PC3|0<<PC4|0<<PC5;
  DDRC  = 1<<PC0|1<<PC1|1<<PC2|1<<PC3|1<<PC4|1<<PC5;

  PORTD = 0<<PD0|0<<PD1|0<<PD2|1<<PD3|0<<PD4|1<<PD5|1<<PD6|1<<PD7;
  DDRD  = 0<<PD0|1<<PD1|0<<PD2|1<<PD3|0<<PD4|1<<PD5|0<<PD6|0<<PD7;

//TIMER1: Mode 0 (Normal), Takt = clk (WGM1x = 0),Input Capture falling edge
  TCCR1A = 0;
  TCCR1B = 0<<ICNC1|0<<ICES1|0<<CS12|0<<CS11|1<<CS10;
  TIMSK1 = 1<<ICIE1|1<<TOIE1;

//TIMER0: Mode 2(CTC): (WGM0x = 0b010, Takt = clk: CS0x = 1), TOP: OCR0A,
//        Toggle OC0B on Compare Match: COM0Bx = 0b01
  TCCR0A = 0<<COM0A1|0<<COM0A0|0<<COM0B1|1<<COM0B0|1<<WGM01|0<<WGM00;
  TCCR0B = 0<<FOC0A |0<<FOC0B |0<<WGM02 |0<<CS02  |0<<CS01 |1<<CS00;
//  TIMSK0 = 0<<OCIE0B|0<<OCIE0A|0<<TOIE0;
//  OCR0A  = 0; //  f = clk/(2*(OCR0A+1))
//  OCR0B  = 0;

//TIMER2: Mode 2(CTC): (WGM2x = 0b010, Takt = clk: CS2x = 1), TOP: OCR2A,
//        Toggle OC2B on Compare Match, COM2Bx = 0b01
  TCCR2A = 0<<COM2A1|0<<COM2A0|0<<COM2B1|1<<COM2B0|1<<WGM21|0<<WGM20;
  TCCR2B = 0<<FOC2A |0<<FOC2B |0<<WGM22 |0<<CS22  |0<<CS21 |1<<CS20;
//  TIMSK2 = 0<<OCIE2B|0<<OCIE2A|0<<TOIE2;
//  OCR2A  = 0; // f = clk/(2*(OCR2A+1))
//  OCR2B  = 0;

// Datenempfang vom GPS-Empfaenger
  UBRR0  = UBRR_VAL;
  UCSR0A = (1<<U2X0);
  UCSR0B = (1<<RXEN0)|(0<<RXCIE0);
  UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);

// Speicherung des DAC-Wertes beim Ausschalten
  EICRA  = (1<<ISC01)|(0<<ISC00);   // Interrupt Request by falling edge of INT0
  EIMSK  = (1<<INT0);               // External Interrupt Request 0 Enable

  sei();
}
///////////////////////// Taktausgaenge konfigurieren /////////////////////////
void takt_adj(void) {
  uint8_t tmp;
  ocr0a    = pgm_read_byte(&freq_p[takt0][0]);
  OCR0A    = ocr0a;
  ocr0b    = pgm_read_byte(&freq_p[takt0][1]);
  OCR0B    = ocr0b;
  TCCR0B   = pgm_read_byte(&freq_p[takt0][2]);
  cnt_max0 = pgm_read_byte(&freq_p[takt0][3]);
  if(cnt_max0) TIMSK0 |= 1<<OCIE0A; else TIMSK0 &= ~(1<<OCIE0A);

  ocr2a    = pgm_read_byte(&freq_p[takt2][0]);
  OCR2A    = ocr2a;
  ocr2b    = pgm_read_byte(&freq_p[takt2][1]);
  OCR2B    = ocr2b;
  tmp      = pgm_read_byte(&freq_p[takt2][2]);
  if(3 == tmp) tmp = 4;
  TCCR2B   = tmp;
  cnt_max2 = pgm_read_byte(&freq_p[takt2][3]);
  if(cnt_max2) TIMSK2 |= 1<<OCIE2A; else TIMSK2 &= ~(1<<OCIE2A);

  lcd_pos(4,1);  lcd_text_pgm(PSTR("T1:    Hz  T2:    Hz"));
  lcd_pos(4,4);  lcd_uint(pgm_read_word(&freq_a[takt0]),3,0,1);
  lcd_data(pgm_read_byte(&freq_p[takt0][4]));   
  lcd_pos(4,15); lcd_uint(pgm_read_word(&freq_a[takt2]),3,0,1); 
  lcd_data(pgm_read_byte(&freq_p[takt2][4]));
}    
////////////// Einstellung der Zeitzone und der Taktausgaenge /////////////////
void menu( void ) {
  int8_t tmp;  
  if(FLAG&T_ZONE) {
  lcd_pos(4,1); lcd_text_pgm(PSTR("      UTC    h?      "));
    time_zone +=  encode_read();
    if(  12 < time_zone) time_zone =  12;
    if( -12 > time_zone) time_zone = -12;
    tmp = time_zone;
    lcd_pos(4,11); 
    if( 0 > tmp) { tmp *=-1; lcd_data('-'); }  else lcd_data('+');
    lcd_uint(tmp , 2 , 0, 1);
    if(get_key_short( 1<<KEY0 )) { 
      FLAG &= ~(MENU|T_ZONE); 
      takt_adj(); 
      encode_read(); 
    }
    return;
  }

  if(FLAG&WAHL) {
    lcd_pos(4,1); lcd_text_pgm(PSTR("   Taktausgang: 1?  "));
    taktq +=  encode_read();
    lcd_pos(4,17);
    if( 0 < taktq) { FLAG |=  TAKTQ; taktq = 1; lcd_data('2'); }
    else           { FLAG &= ~TAKTQ; taktq = 0; lcd_data('1'); }
    if(get_key_short( 1<<KEY0 )) { FLAG &= ~WAHL; encode_read(); }
    return; 
  }

  lcd_pos(4,1);  lcd_text_pgm(PSTR("T1:    Hz T2:    Hz "));
  if(FLAG&TAKTQ) { 
    lcd_pos(4,20);lcd_data('?');
    takt2 +=encode_read(); 
    if(0 > takt2) takt2 = 0;
    if(28 < takt2) takt2 = 28;
  }
  else {
    lcd_pos(4,10);lcd_data('?');
    takt0 +=encode_read(); 
    if(0 > takt0) takt0 = 0;
    if(28 < takt0) takt0 = 28;
  }  
  lcd_pos(4,4);  lcd_uint(pgm_read_word(&freq_a[takt0]),3,0,1);
  lcd_data(pgm_read_byte(&freq_p[takt0][4]));   
  lcd_pos(4,14); lcd_uint(pgm_read_word(&freq_a[takt2]),3,0,1); 
  lcd_data(pgm_read_byte(&freq_p[takt2][4]));

  if(get_key_short( 1<<KEY0 )) { FLAG &= ~MENU; takt_adj();}
  if(get_key_long( 1<<KEY0 )) { FLAG &= ~MENU; takt_adj();reset = 1;}
}
////////////////// Dekodierung des GGA-Strings im GPS-Signal //////////////////
/* GGA message-structure:
$GPGGA,hhmmss.ss,Latitude,N,Longitude,E,FS,NoSV,HDOP,msl,m,Altref,m,
                                                 DiffAge,DiffStation*cs<CR><LF>
Example:
$GPGGA,092725.00,4717.11399,N,00833.91590,E,1,8,1.01,499.6,M,48.0,M,,0*5B
Field Example  Format      Name      Unit    Description
No.                                       
 0 $GPGGA      string      $GPGGA      - Message ID, GGA protocol header
 1 092725.00   hhmmss.sss  hhmmss.ss   - UTC Time, Current time
 2 4717.11399  ddmm.mmmm   Latitude    - Latitude, Degrees + minutes
 3 N           character   N           - N/S Indicator, N=north or S=south
 4 00833.91590 dddmm.mmmm  Longitude   - Longitude,Degrees+minutes
 5 E           character   E           - E/W-indicator,E=east or W=west
 6 1           digit       FS          - Position Fix Status Indicator
 7 8           numeric     NoSV        - Satellites Used, Range 0 to 12
 8 1.01        numeric     HDOP        - HDOP, Horizontal Dilution of Precision
 9 499.6       numeric     msl         m MSL Altitude
10 M           character   uMsl        - Units, Meters (fixed field)
11 48.0        numeric     Altref      m Geoid Separation
12 M           character   uSep        - Units, Meters (fixed field)
13 -           numeric     DiffAge     s Age of Differential Corrections,
                                         Blank (Null) fields when DGPS is not used
14 0           numeric     DiffStation - Diff. Reference Station ID
15 *5B         hexadecimal cs          - Checksum
16 -           character   <CR><LF>    - Carriage Return and Line Feed
*/
void gga_rx( void ) {
  uint8_t rx = 0, i = 0, rx_gga = 0;
  int8_t tmp1;
  while(16>rx_gga) {
    while(!(UCSR0A&(1<<RXC0))) {if(FLAG&TIMEPULSE) return;} // Wait for rx-data
    rx  = UDR0;
    switch(rx_gga) {
      case  0: if(rx == '$') rx_gga++;                  break;
      case  1: if(rx == 'G') rx_gga++; else rx_gga = 0; break;
      case  2: if(rx == 'P') rx_gga++; else rx_gga = 0; break;
      case  3: if(rx == 'G') rx_gga++; else rx_gga = 0; break;
      case  4: if(rx == 'G') rx_gga++; else rx_gga = 0; break;
      case  5: if(rx == 'A') rx_gga++; else rx_gga = 0; break;
      case  6: if(rx == ',') rx_gga++; else rx_gga = 0; break; 
      case  7: if(rx == ',') { rx_gga++; i=0;break;}
               if(rx == '.') break;
               if(6>i) { gga_time[i] = rx; i++; } break;
      case  8: if(rx == ',') { rx_gga++; i=0; break;}
               if(10>i) { gga_latitude[i] = rx; i++; }
               break;
      case  9: if(rx == ',') { rx_gga++; i=0; break;}
               if(!i) { gga_n = rx; i++; }
               break;
      case 10: if(rx == ',') { rx_gga++; i=0; break;}
               if(11>i) { gga_longitude[i] = rx; i++; }
               break;
      case 11: if(rx == ',') { rx_gga++; i=0; break;}
               if(!i) { gga_e = rx; i++; }
               break;
      case 12: if(rx == ',') { rx_gga++; i=0; break;}
               if(!i) { gga_fs = rx; i++; }
               break;
      case 13: if(rx == ',') { rx_gga++; i=0; break;}
               if(2>i) { gga_nosv[i] = rx; i++; }
               break;
      case 14: if(rx == ',') { rx_gga++; i=0; break;} // HDOP
               break;
      case 15: if(rx == '.') { gga_msl[i] = 'm'; rx_gga++;  break;}
               if(4>i) { gga_msl[i] = rx; i++; }
    }
  }
  if(FLAG&MENU) menu(); // Menuaufruf bei Einstellmodus 1x pro Sekunde von hier
  if(('1' == gga_fs)||('2' == gga_fs)) {        // Fix Status
    lcd_pos(1,1);
    // GPS_POSITION 20Stellen
    lcd_data(' ');
    lcd_data(gga_latitude[0]);
    lcd_data(gga_latitude[1]);
    lcd_data(0xDF);     // 
    for(i=2; i<10; i++) lcd_data(gga_latitude[i]);
    lcd_data(gga_n);
    lcd_data(' ');
    lcd_data(' ');
    //GPS_MSL 8 Stellen
    tmp1=0;
    for(i=0; i<5; i++) {
      if(tmp1) lcd_data(' ');
      else {
        lcd_data(gga_msl[i]); 
        if(gga_msl[i] == 'm') {  tmp1++; }
      }
    }
      lcd_pos(2,1);
    for(i=0; i<3; i++) lcd_data(gga_longitude[i]);
    lcd_data(0xDF);     // 
    for(i=3; i<11; i++) lcd_data(gga_longitude[i]);
    lcd_data(gga_e);


    // GPS_SATS 2Stellen
    lcd_text_pgm(PSTR(" SAT:"));
    if(gga_nosv[0] > '1') {lcd_data('0'); lcd_data(gga_nosv[0]);} 
    else {lcd_data(gga_nosv[0]); lcd_data(gga_nosv[1]);  }


    // GPS_TIME 8 Stellen
    lcd_pos(3,1);
    tmp1 = (((gga_time[0] -0x30) * 10) +(gga_time[1] - 0x30)) + time_zone;
    if(tmp1 <  0) tmp1 += 24;
    if(tmp1 > 23) tmp1 -= 24;
    lcd_uint(tmp1 , 2, 0, 1 ); 
    lcd_data(':');
    lcd_data(gga_time[2]);
    lcd_data(gga_time[3]);
    lcd_data(':');
    lcd_data(gga_time[4]);
    lcd_data(gga_time[5]);

  }   
}
/////////////////////////////////// MAIN //////////////////////////////////////
int main(void) {

  int16_t drift;
  static uint16_t icr_alt1, imp_dac, delta ,icr_alt10;
  static uint8_t icr_count;
  static uint32_t imp;

  cfg();
  lcd_init();
  encode_init();
  SPI_MasterInit();
  read_EEPROM();
  FLAG |=  NO_GPS;
  takt_adj();
  while(1) {
    while(!(FLAG&TIMEPULSE)) {     // GPS-Empfang pruefen
      if(FLAG&NO_GPS) {
        gga_fs = 0;
        lcd_clear();
        write_DAC();
        FLAG &= ~IN_SYNC;
// Warten auf ersten Sekundenimpuls
        while(!(FLAG&TIMEPULSE)) {
          if(!(FLAG&MENU)) {
           // Aufruf Einstellmodus fuer Taktausgaenge
              if(get_key_short( 1<<KEY0 )) { FLAG |= MENU|WAHL;   encode_read(); }
           // Aufruf Einstellmodus fuer Zeitzone
              if(get_key_long( 1<<KEY0 ))  { FLAG |= MENU|T_ZONE; encode_read(); }
            }
            else  menu();
            _delay_ms(300);
          }
        icr_alt1 = icr;           // Sekundenimpuls eingetroffen
        FLAG &= ~NO_GPS;
        FLAG &= ~TIMEPULSE;
        icr_count = 60;           // Synchronisationszeit
      }
    }
    FLAG &= ~TIMEPULSE;
// Taktdifferenz pro Sekunde, Achtung! Rechnung mit unsigned 16bit
// 0x0000 - 0x0001 = 0xFFFF und 0xFFFF + 0x0001 = 0x0000 !!!!!
    drift = icr - icr_alt1 - REST1;
    icr_alt1 = icr;
     // Regelung erst bei stabilem Empfang (FIX-STATUS > 0):
    if(('1' == gga_fs)||('2' == gga_fs))  icr_count--;
    if(!(FLAG&IN_SYNC)){
      if(55 > icr_count) {
        dac_wert -= 3*drift;  // DAC-Wert anpassen
        if((drift > 1) || (drift < -1))  dac_wert -= (icr_count>>1)*drift;
        write_DAC();
      }
      lcd_pos(3,14);
      lcd_text_pgm(PSTR(" SYN"));
      lcd_uint(icr_count , 2 , 0, 1 );
      lcd_data('s');
    }
    if(icr_count == 0) {      // Nach 60s Synchronisationszeit
      FLAG |= IN_SYNC;
      icr_alt10 = icr;        // Jetzt Taktdifferenz pro 10Sekunden
      icr_count = 11;
      lcd_pos(3,14);
      lcd_text_pgm(PSTR("    "));
    }
    if(((drift>1) || (drift<-1)) && (FLAG&IN_SYNC)){ // Taktdifferenz zu gross
      FLAG &= ~IN_SYNC;
      dac_wert -= drift;
      write_DAC();
      icr_count = 60;
     }
     if(!(FLAG&OLD_SYNC_VAL) && (FLAG&IN_SYNC)) { // Beginn: In Sync
      FLAG |= OLD_SYNC_VAL;
      imp_dac = 0;      // Anzahl der DAC-Wert Aenderungen
      imp = 0;          // Anzahl der Messintervalle
    }
    if((FLAG&OLD_SYNC_VAL) && !(FLAG&IN_SYNC)) { // Beginn: Async
      FLAG &= ~OLD_SYNC_VAL;
      delta++;
    }
    if(FLAG&IN_SYNC) {             // Synchroner Betrieb
      lcd_pos(3,18);
      if(delta) lcd_data('!'); 
      else      lcd_data(' ');
      lcd_uint(delta , 2 , 0, 1 ); // Anzahl Synchronausfaelle
    }
    if((FLAG&IN_SYNC) && (icr_count==1)) { // 10s-Messintervall
      icr_count= 11;
      imp++;                              // Anzahl der Messintervalle
// Taktdifferenz pro 10Sekunden, Achtung! Rechnung mit unsigned 16bit
// 0x0000 - 0x0001 = 0xFFFF und 0xFFFF + 0x0001 = 0x0000 !!!!!
      drift =  icr - icr_alt10 - REST10;
      icr_alt10 = icr;
      if(drift) {                         // DAC einstellen
        dac_wert -= drift;
        write_DAC();
        imp_dac++;
      }
      lcd_pos(3,14);
      lcd_data(0xFD);     // :-
      if(!imp_dac) imp_dac = 1;               // Durchschnittliches Intervall
      lcd_uint((imp*10)/imp_dac , 2 , 0, 1 ); // der DAC-Wert Aenderungen
      lcd_data('s');
    }
    if(!(FLAG&MENU)) {
      // Aufruf Einstellmodus fuer Taktausgaenge
      if(get_key_short( 1<<KEY0 )) { FLAG |= MENU|WAHL;   encode_read(); }
      // Aufruf Einstellmodus fuer Zeitzone
      if(get_key_long( 1<<KEY0 ))  { FLAG |= MENU|T_ZONE; encode_read(); }
    }
    else menu(); // Menuaufruf bei Einstellmodus 1x pro Sekunde von hier
    if(!(FLAG&NO_GPS)) gga_rx();          // und 1x pro Sekunde von hier
    if(reset) { 
      reset = 0;
      imp_dac = 0;      // Anzahl der DAC-Wert Aenderungen
      imp = 0;          // Anzahl der Messintervalle      
      delta = 0;        // Anzahl der Synchronausfaelle
    } 
  }
}
/////////////// Zaehlung: Takt/2e16, Abfrage Encoder und Taste ////////////////
ISR(TIMER1_OVF_vect) {     // 305 oder 306x pro Sek. (ca. alle 3,3ms )
  static uint8_t  tic, ct0, ct1, rpt;
  uint8_t         i;
  int8_t          neu, diff;
  tic++;
  t1_ovf++;
  if(t1_ovf>306) { FLAG |=  NO_GPS; FLAG &= ~TIMEPULSE; t1_ovf = 0; }

// Ca. alle 3,3ms Encoder auslesen:
    neu = 0;
    if( PHASE_A )    neu  = 3;
    if( PHASE_B )    neu ^= 1;           // convert gray to binary
    diff = last - neu;                   // difference last - neu
    if( diff & 1 ) {                     // bit 0 = value (1)
      last = neu;                        // store neu as next last
      enc_delta += (diff & 2) - 1;       // bit 1 = direction (+/-)
    }
// Ca. alle 20ms Tastenentprellung:
  if(tic > 5){                                  
    tic = 0;
    i = key_state ^ ~KEY_PIN;            // key changed ?
    ct0 = ~( ct0 & i );                  // reset or count ct0
    ct1 = ct0 ^ (ct1 & i);               // reset or count ct1
    i &= ct0 & ct1;                      // count until roll over ?
    key_state ^= i;                      // then toggle debounced state
    key_press |= key_state & i;          // 0->1: key press detect
    if( (key_state & REPEAT_MASK) == 0 ) // check repeat function
      rpt = REPEAT_START;                // start delay
    if( --rpt == 0 ) key_rpt |= key_state & REPEAT_MASK;
  }
}
// Bei 305/306 Timer1-Ueberlaeufen und Sekundenimpuls Zaehlerstand speichern //
ISR(TIMER1_CAPT_vect) {           // 1x pro Sekunde
  icr = ICR1;
  FLAG |= TIMEPULSE;
//  if(('1' == gga_fs)||('2' == gga_fs)) FLAG |= TIMEPULSE;
  if(t1_ovf<305) { FLAG |=  NO_GPS; FLAG &= ~TIMEPULSE; }
  t1_ovf = 0;
}
///////////////////////////////// Taktausgang 1 ///////////////////////////////
ISR(TIMER0_COMPA_vect) {          // Abhaengig von eingestellter Frequenz
  static uint8_t cnt;
  if(!cnt) { OCR0A=ocr0a; }
  cnt++;
  if(cnt_max0==cnt) { OCR0A=ocr0b; cnt=0; }
}
///////////////////////////////// Taktausgang 2 ///////////////////////////////
ISR(TIMER2_COMPA_vect) {          // Abhaengig von eingestellter Frequenz
  static uint8_t cnt;
  if(!cnt) { OCR2A=ocr2a; }
  cnt++;
  if(cnt_max2==cnt) { OCR2A=ocr2b; cnt=0;}
}
//////////////// Speichern des DAC-Wertes bei Stromausfall ////////////////////
ISR(INT0_vect) { write_EEPROM(); }
